from typing import Union, Optional


class Graph:
    def __init__(self, nodes: list[list[Union[float, int]]], edge_index: Optional[list[list[int]]],
                 edge_attr: Optional[list[list[float]]], pos: Optional[list[list[float]]],
                 labels: list[list[float]], angles = None
                 ) -> None:
        self.nodes = nodes
        self.edge_index = edge_index
        self.edge_attr = edge_attr
        self.pos = pos
        self.labels = labels
        self.angles = angles



from laegr.featurization.Graph import Graph
from ovito.io import import_file
from ovito.data import NearestNeighborFinder, CutoffNeighborFinder
from typing import Union, Any, Optional
import ovito.data
from abc import ABC, abstractmethod
import math
import numpy as np

class GetGraph(ABC):
    """
    Base class to generate the Graph (node features, edges and labels/PE)
    """
    def __init__(self, data: Union[str, ovito.data.DataCollection],
                 labels: Optional[list[list[float]]]) -> None:
        self.labels = labels
        if isinstance(data, str):
            # assert labels is None, "The labels should be None if the input data is lammps file."
            pipeline = import_file(data)
            self.data = pipeline.compute()
        else:
            self.data = data
            # self.labels = labels

    @abstractmethod
    def getNodeFeatures(self) -> list[list[Union[float, int]]]:
        pass

    def getLabels(self) -> list[list[float]]:
        if self.labels:
            return self.labels
        labels = self.data.particles['c_eng'][...]
        labels = [[i] for i in labels]
        return labels


class GraphFeatures(GetGraph):
    def __init__(self, data: Union[str, ovito.data.DataCollection],
                 labels: Optional[list[list[float]]]) -> None:
        super().__init__(data, labels)

    def toGraph(self, criteria: Union[int, float]) -> Graph:
        nodes = self.getNodeFeatures()
        edge_index, edge_attr = self.getEdge(criteria)
        labels = self.getLabels()
        graph_features = Graph(nodes, edge_index, edge_attr, None, labels)
        return graph_features

    def getNodeFeatures(self) -> list[list[Union[float, int]]]:
        raise NotImplementedError

    def getEdge(self, criteria: Union[int, float]) -> tuple[list[list[int | Any]], list[list[float]]]:
        self.finder = None
        if isinstance(criteria, int):
            self.finder = NearestNeighborFinder(criteria, self.data)
        if isinstance(criteria, float):
            self.finder = CutoffNeighborFinder(criteria, self.data)
        assert self.finder is not None, "The criteria to find neighbor should be int or float."
        edge_index = []
        edge_attr = []
        for index in range(self.data.particles.count):
            for neigh in self.finder.find(index):
                edge_index.append([index, neigh.index])
                edge_attr.append(self.getEdgeAttribute(neigh))
        return edge_index, edge_attr

    @staticmethod
    @abstractmethod
    def getEdgeAttribute(neigh: NearestNeighborFinder) -> list[float]:
        pass


class AtomicNumber(GraphFeatures):
    def getNodeFeatures(self) -> list[list[int]]:
        ptypes = self.data.particles.particle_types
        nodes = []
        transform = {1: 42, 2: 41, 3: 73, 4: 74}
        for atom_type in ptypes:
            cur_node = transform[atom_type]
            nodes.append(cur_node)
        return nodes

    @staticmethod
    @abstractmethod
    def getEdgeAttribute(neigh: NearestNeighborFinder.Neighbor) -> list[float]:
        pass



class AtomicNumberDistance(AtomicNumber):
    @staticmethod
    def getEdgeAttribute(neigh: NearestNeighborFinder.Neighbor) -> list[float]:
        return [neigh.distance]



class AtomicNumberPos(AtomicNumberDistance):
    def toGraph(self, criteria) -> Graph:
        nodes = self.getNodeFeatures()
        edge_index, edge_attr = self.getEdge(criteria)
        pos = self.getPosition()
        labels = self.getLabels()
        angles = self.getAngle()
        graph_features = Graph(nodes, edge_index, edge_attr, pos, labels, angles)
        return graph_features
    

    def getPosition(self) -> list[list[float]]:
        pos_list = []
        for cur_pos in self.data.particles.positions:
            pos_list.append(list(cur_pos))
        return pos_list
    
    def getAngle(self) -> list[float]:
        angles = []
        for index in range(self.data.particles.count):
            cur_distance = []
            for neigh in self.finder.find(index):
                cur_distance.append(neigh.delta)
            for idx_1 in range(len(cur_distance)):
                for idx_2 in range(len(cur_distance)):
                    if idx_1 == idx_2:
                        continue
                    cur_angle = self.__calcAnge(cur_distance[idx_1], cur_distance[idx_2])
                    angles.append(cur_angle)
        return angles
    
    @staticmethod
    def __calcAnge(x: list[float], y: list[float]) -> float:
        dot_product = np.dot(x, y)
        m_x = np.linalg.norm(x)
        m_y = np.linalg.norm(y)
        val = dot_product / (m_x * m_y)
        if val < -1:
            val = -1
        if val > 1:
            val = 1
        angle = math.acos(val)
        return angle

class AtomicMass(GraphFeatures):
    def getNodeFeatures(self) -> list[list[float]]:
        nodes = self.data.particles['Mass'][...]
        nodes = [[i] for i in nodes]
        return nodes

    @staticmethod
    @abstractmethod
    def getEdgeAttribute(neigh: NearestNeighborFinder.Neighbor) -> list[float]:
        pass



class AtomicMassDistance(AtomicMass):
    @staticmethod
    def getEdgeAttribute(neigh: NearestNeighborFinder.Neighbor) -> list[float]:
        return [neigh.distance]



class AtomicMass3D(AtomicMass):
    @staticmethod
    def getEdgeAttribute(neigh: NearestNeighborFinder.Neighbor) -> list[float]:
        return list(neigh.delta)



class OnehotGraph(GraphFeatures):
    def getNodeFeatures(self) -> list[list[int]]:
        ptypes = self.data.particles.particle_types
        nodes = []
        for atom_type in ptypes:
            cur_node = self.toOneHot(atom_type)
            nodes.append(cur_node)
        return nodes

    @staticmethod
    def toOneHot(atom_type: int) -> list[int]:
        if atom_type == 1:
            new_atom_type = 2
        elif atom_type == 2:
            new_atom_type = 1
        else:
            new_atom_type = atom_type
        feature = [0, 0, 0, 0]
        feature[new_atom_type - 1] = 1
        return feature

    @staticmethod
    @abstractmethod
    def getEdgeAttribute(neigh: NearestNeighborFinder.Neighbor) -> list[float]:
        pass



class OnehotDistance(OnehotGraph):
    @staticmethod
    def getEdgeAttribute(neigh: NearestNeighborFinder.Neighbor) -> list[float]:
        return [neigh.distance]



class Onehot3D(OnehotGraph):
    @staticmethod
    def getEdgeAttribute(neigh: NearestNeighborFinder.Neighbor) -> list[float]:
        return list(neigh.delta)








import warnings
warnings.filterwarnings('ignore', message='.*OVITO.*PyPI')

from laegr.featurization import GNNData
from laegr.model import Train, TrainEdge, TrainSchNet,  TrainCrystal
from laegr.model import LoadData



import torch
from torch_geometric.nn import SchNet, CGConv, global_mean_pool


class Model(torch.nn.Module):
    def __init__(self, gnn_model: torch.nn, output_model: torch.nn, activation: torch.nn.ReLU()) -> None:
        super(Model, self).__init__()
        self.gnn = gnn_model
        self.output = output_model
        self.activation = activation

    def forward(self, x: torch.Tensor, edge_index: torch.Tensor, edge_attr: torch.Tensor) -> torch.Tensor:
        for idx in range(len(self.gnn)-1):
            x = self.gnn[idx](x, edge_index, edge_attr)
            x = self.activation(x)
        x = self.gnn[-1](x, edge_index, edge_attr)
        out = self.output(x)
        return out
    # def forward(self, x: torch.Tensor, edge_index: torch.Tensor, edge_attr: torch.Tensor) -> torch.Tensor:
    #     for idx in range(len(self.gnn)):
    #         x = self.gnn[idx](x, edge_index, edge_attr)
    #         x = self.activation(x)
    #     # x = self.gnn[-1](x, edge_index, edge_attr)
    #     out = self.output(x)
    #     return out


class Model_SchNet(Model):
    def forward(self, x: torch.Tensor, pos: torch.Tensor, batch: torch.Tensor) -> torch.Tensor:
        for idx in range(len(self.gnn)-1):
            x = self.gnn[idx](x, pos, batch)
            x = self.activation(x)
        x = self.gnn[-1](x, pos, batch)
        out = self.output(x)
        return out


class Model_GNN_Base():
    def __init__(self, first_layer_size: int, embedding_size: int, layer_type: torch.nn, layer_number: int) -> None:
        self.embedding_size = embedding_size
        self.layer_type = layer_type
        self.layer_number = layer_number
        self.first_layer_size = first_layer_size

    def build_model(self) -> torch.nn.ModuleList:
        model = torch.nn.ModuleList()
        for idx in range(self.layer_number):
            if (idx == 0):
                cur_layer = self.define_layer(self.first_layer_size)
            else:
                cur_layer = self.define_layer(self.embedding_size)
            model.append(cur_layer)
        return model

    def define_layer(self, input_size: int):
        return self.layer_type(input_size, self.embedding_size)



class Model_GNN_Crystal():
    def __init__(self, first_layer_size: int, layer_number: int, edge_dim: int) -> None:
        self.layer_number = layer_number
        self.first_layer_size = first_layer_size
        self.edge_dim = edge_dim

    def build_model(self) -> torch.nn.ModuleList:
        model = torch.nn.ModuleList()
        for _ in range(self.layer_number):
            cur_layer = CGConv(self.first_layer_size, dim=self.edge_dim)
            model.append(cur_layer)
        return model
        


class Model_GNN_SchNet():
    def __init__(self, layer_number: int = 1) -> None:
        self.layer_number = layer_number

    def build_model(self,   hidden_channels: int = 128, num_filters: int = 128, num_interactions: int = 6, 
                            num_gaussians: int = 50,    cutoff: float = 4.7,   max_num_neighbors: int = 26
                    ) -> torch.nn.ModuleList:
        model = torch.nn.ModuleList()
        for _ in range(self.layer_number):
            cur_layer = SchNet(hidden_channels=hidden_channels, num_filters=num_filters, num_interactions=num_interactions,
                                num_gaussians=num_gaussians, cutoff=cutoff, max_num_neighbors=max_num_neighbors)
            model.append(cur_layer)
        return model

    # def define_layer(self):
    #     return self.layer_type()


class Model_GNN_NN(Model_GNN_Base):
    def __init__(self, first_layer_size: int, embedding_size: int, layer_type: torch.nn, layer_number: int, edge_model_list: torch.nn) -> None:
        super().__init__(first_layer_size, embedding_size, layer_type, layer_number)
        self.edge_model_list = edge_model_list
    
    def build_model(self) -> torch.nn.ModuleList:
        model = torch.nn.ModuleList()
        for idx in range(self.layer_number):
            if (idx == 0):
                cur_layer = self.define_layer(self.first_layer_size, idx)
            else:
                cur_layer = self.define_layer(self.embedding_size, idx)
            model.append(cur_layer)
        return model
    
    def define_layer(self, input_size: int, idx: int):
        return self.layer_type(input_size, self.embedding_size, self.edge_model_list[idx])



        




from torch_geometric.data.data import Data
import torch
import ovito.data
from typing import Union, Optional

from laegr.featurization.Graph import Graph
from laegr.featurization.GraphFeatures import AtomicMassDistance, AtomicMass3D, AtomicNumberDistance, AtomicNumberPos
from laegr.featurization.GraphFeatures import OnehotDistance, Onehot3D


class GNNData:
    def __init__(self, source: Union[str, ovito.data.DataCollection], criteria: Union[int, float], feature: str,
                 label: Optional[list[list[float]]] = None) -> None:
        self.data = source
        self.label = label
        self.criteria = criteria
        self.feature = feature

    def toData(self) -> Data:
        graph = self.toGraph()
        data = self.toPytorchGNNData(graph)
        return data

    def toGraph(self) -> Graph:
        match self.feature:
            case "AtomicMassDistance":
                graph = AtomicMassDistance(self.data, self.label).toGraph(self.criteria)
            case "AtomicMass3D":
                graph = AtomicMass3D(self.data, self.label).toGraph(self.criteria)
            case "AtomicNumberDistance":
                graph = AtomicNumberDistance(self.data, self.label).toGraph(self.criteria)
            case "AtomicNumberPos":
                graph = AtomicNumberPos(self.data, self.label).toGraph(self.criteria)
            case "OnehotDistance":
                graph = OnehotDistance(self.data, self.label).toGraph(self.criteria)
            case "Onehot3D":
                graph = Onehot3D(self.data, self.label).toGraph(self.criteria)
            case _:
                raise Exception('''Only support six different features:   
                                        AtomicMassDistance,             AtomicMass3D,
                                        AtomicNumberDistance,           AtomicNumberPos,
                                        OnehotDistance,                 Onehot3D.''')
        return graph

    @staticmethod
    def toPytorchGNNData(graph: Graph) -> Data:
        labels = graph.labels
        if graph.labels:
            labels = torch.tensor(graph.labels, dtype=torch.float)
        if graph.pos:
            nodes = torch.tensor(graph.nodes, dtype=torch.long)
            edge_index = torch.tensor(graph.edge_index, dtype=torch.long)
            edge_pos = torch.tensor(graph.pos, dtype=torch.float)
            edge_attr = torch.tensor(graph.edge_attr, dtype=torch.float)
            angles = torch.tensor(graph.angles, dtype=torch.float)
            data = Data(z=nodes, edge_index=edge_index.t().contiguous(), edge_attr=edge_attr, pos=edge_pos, y=labels, angels=angles)
        else:
            nodes = torch.tensor(graph.nodes, dtype=torch.float)
            edge_index = torch.tensor(graph.edge_index, dtype=torch.long)
            edge_attr = torch.tensor(graph.edge_attr, dtype=torch.float)
            data = Data(x=nodes, edge_index=edge_index.t().contiguous(), edge_attr=edge_attr, y=labels)
        return data



from torch_geometric.nn import NNConv, MessagePassing
from torch_geometric.typing import Adj, OptPairTensor, OptTensor, Size
from typing import Union, Tuple
import torch
from torch_geometric.nn.inits import reset


class NodeCat(MessagePassing):
    def __init__(self, in_channels: int, out_channels: int, nn: torch.nn, aggr: str = "add", num_neighbor: int = 26, **kwargs):
        super().__init__(aggr, **kwargs)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.nn = nn
        self.num_neighbor = num_neighbor

        if isinstance(in_channels, int):
            in_channels = (in_channels, in_channels)
        self.in_channels_l = in_channels[0]
        self.reset_parameters()

    def reset_parameters(self):
        reset(self.nn)
    
    def forward(self, x: Union[torch.Tensor, OptPairTensor], edge_index: Adj, edge_attr: OptTensor = None, size: Size = None) -> torch.Tensor:
        if isinstance(x, torch.Tensor):
            x: OptPairTensor = (x, x)

        # propagate_type: (x: OptPairTensor, edge_attr: OptTensor)
        out = self.propagate(edge_index, x=x, edge_attr=edge_attr, size=size)
        x_r = x[1]
        out = torch.cat((x_r, out), 1)
        return out
    
    def aggregate(self, inputs: torch.Tensor) -> torch.Tensor:
        out = inputs.view(-1, self.num_neighbor*self.out_channels)
        return out
    
    def message(self, x:torch.tensor, edge_attr: torch.Tensor, index: torch.Tensor) -> torch.Tensor:
        x_j = torch.clone(x[1])
        weight = self.nn(edge_attr)
        x_j = x_j[index]
        weight = weight.view(-1, self.in_channels_l, self.out_channels)
        return torch.matmul(x_j.unsqueeze(1), weight).squeeze(1)



class NNConv_Edge_Interaction(NNConv):
    def __init__(self, in_channels: Union[int, Tuple[int, int]], out_channels: int, nn, nn_2, aggr: str = 'add', root_weight: bool = True, bias: bool = True, **kwargs):
        super().__init__(in_channels, out_channels, nn, aggr, root_weight, bias, **kwargs)
        self.nn_2 = nn_2
        
    def message(self, x_j: torch.Tensor, edge_attr: torch.Tensor) -> torch.Tensor:
        weight_1 = self.nn(edge_attr)
        weight_1 = weight_1.view(-1, self.in_channels_l, self.out_channels)
        print(x_j)
        print(x_j.shape)
        print(edge_attr)
        print(edge_attr.shape)
        edge_attr_new = edge_attr.unsqueeze(1)
        edge_attr_new = (edge_attr_new + torch.transpose(edge_attr_new, 0, 1))
        edge_attr_new = edge_attr_new.view(-1, 3)
        print(edge_attr_new.shape)
        weight_2 = self.nn_2(edge_attr_new)
        weight_2 = weight_2.view(-1, self.in_channels_l, self.out_channels)
        x_j_new = x_j.unsqueeze(1)
        x_j_new = (x_j_new + torch.transpose(x_j_new, 0, 1))
        x_j_new = torch.matmul(x_j_new, weight_2)
        x_j_new = torch.sum(x_j_new, dim=0)
        print(x_j_new.shape)
        return torch.matmul(x_j.unsqueeze(1), weight_1).squeeze(1) + x_j_new


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="laegr",                            
    version="1.1",                            
    author="Yi Yao, Lin Li",
    author_email="yyao26@crimson.ua.edu",                        
    description="Python package to extract graph structure of local atomic environment and build GNN to predict atomic properties (Potential Energy)",
    long_description=long_description,          
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),        
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],                                         
    python_requires='>=3.8',                  
    py_modules=["laegr"],                   
    # package_dir={'':'LAEGR/src'},            
    install_requires=[]                       
)


import numpy as np
import pickle
import torch
from typing import Union, List, Tuple, Optional
from torch_geometric.data.data import Data
from torch.utils.data import TensorDataset, DataLoader
from sklearn.model_selection import KFold

# class LoadData():
#     def __init__(self, path: Union[str, list[str]]):
#         self.__List = self.__readList(path)

#     @staticmethod
#     def __readList(path: Union[str, list[str]]):
#         print("Start reading Data")
#         listRawData = []
#         if isinstance(path, str):
#             print("single data")
#             open_file   = open(path, "rb")
#             listRawData = pickle.load(open_file)
#             open_file.close()
#         elif isinstance(path, list):
#             print("Multiple data.")
#             for cur_path in path:
#                 print("Read current data ...")
#                 cur_file = open(cur_path, 'rb')
#                 cur_data = pickle.load(cur_file)
#                 cur_file.close()
#                 listRawData.extend(cur_data)
#         else:
#             raise Exception("Please provide correct data path!")
#         print("Reading data Done")
#         return listRawData
    
#     def transformInput(self, partition: float = 0.8, train_batch_size: int = 1, val_batch_size: int = 1, shuffle: bool = True, random_see: int = 66) -> Tuple[DataLoader, Optional[DataLoader]]:
#         data_size    = len(self.__List)
#         random.seed(random_see)
#         if (shuffle):
#             random.shuffle(self.__List)
#         if partition == -1:
#             train_loader = DataLoader(self.__List, batch_size=train_batch_size)
#             return train_loader, None
#         train_loader       = DataLoader(self.__List[:int(data_size*partition)], batch_size=train_batch_size)
#         validation_loader  = DataLoader(self.__List[int(data_size*partition):], batch_size=val_batch_size)
#         return train_loader, validation_loader



class LoadData:
    def __init__(self, data_src: str) -> None:
        self.data_type = False
        self.data = self.readData(data_src)

    def KFold(self, fold=10, batch_size=1, random_seed=7268128) -> List[Tuple[DataLoader]]:
        kFold_data = []
        kf = KFold(n_splits=fold, shuffle=True, random_state=random_seed)
        for (train_index, test_index) in kf.split(self.data):
            if self.data_type == 'np':
                cur_train_fold = self.np2DataLoader(self.data[train_index], batch_size)
                cur_test_fold  = self.np2DataLoader(self.data[test_index], batch_size)
            else:
                cur_train_fold = DataLoader(self.data, sampler=train_index, batch_size=batch_size, shuffle=False)
                cur_test_fold  = DataLoader(self.data, sampler=test_index, batch_size=batch_size, shuffle=False)
            kFold_data.append((cur_train_fold, cur_test_fold))
        return kFold_data
    
    @staticmethod
    def np2DataLoader(data: np.ndarray, batch_size) -> DataLoader:
        x, y = torch.Tensor(data[:, :-1]), torch.Tensor(data[:, -1:])
        dataset = TensorDataset(x, y)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=False)
        return dataloader

    def readData(self, data_src: str) -> Union[np.ndarray, List[Data]]:
        print("Read data!")
        if "npy" in data_src:
            print("np_ndarray data format")
            self.data_type = 'np'
            data = self.load_NP_ARRAY(data_src)
        else:
            print("list of gnn data format")
            self.data_type = 'gnn'
            data = self.load_Pickled_List(data_src)
        return data
    
    @staticmethod
    def load_NP_ARRAY(data_src: str) -> np.ndarray:
        data = np.load(data_src)
        return data
    
    @staticmethod
    def load_Pickled_List(data_src: str) -> List[Data]:
        open_file   = open(data_src, "rb")
        data = pickle.load(open_file)
        open_file.close()
        return data
    
    def transformInput(self, partition: float = 0.8, train_batch_size: int = 1, val_batch_size: int = 1, shuffle: bool = True, random_see: int = 66) -> Tuple[DataLoader, Optional[DataLoader]]:
        data_size    = len(self.data)
        random.seed(random_see)
        if (shuffle):
            random.shuffle(self.data)
        if partition == -1:
            train_loader = DataLoader(self.data, batch_size=train_batch_size)
            return train_loader, None
        train_loader       = DataLoader(self.data[:int(data_size*partition)], batch_size=train_batch_size)
        validation_loader  = DataLoader(self.data[int(data_size*partition):], batch_size=val_batch_size)
        return train_loader, validation_loader